import os

import numpy as np
import torch
import hydra
from hydra.utils import instantiate
from omegaconf import OmegaConf, DictConfig
import random

from torch_mist.estimators.discriminative import DiscriminativeMIEstimator
from torch_mist.estimators.hybrid import HybridMIEstimator
from torch_mist.utils import train_mi_estimator, evaluate_mi
from torch_mist.utils.estimation import _instantiate_estimator, estimate_mi
from torch_mist.utils.logging.metrics import compute_mean_std


@hydra.main(
    config_path="config", config_name="config.yaml", version_base="1.1"
)
def parse(conf: DictConfig):
    # Set the random seed
    if "seed" in conf:
        torch.manual_seed(conf.seed)
        np.random.seed(conf.seed)
        random.seed(conf.seed)

        if hasattr(conf.device, "tensor_cores"):
            torch.set_float32_matmul_precision(conf.device.matmul_precision)

    # Change the location to the original working directory (to fix issues with relative paths)
    hydra_wd = os.getcwd()
    os.chdir(hydra.utils.get_original_cwd())

    # Instantiate the parameters and metadata if any
    if "params" in conf:
        conf.params = instantiate(conf.params)

    # Instantiate the loggers
    print("Instantiating the logger")
    logger = instantiate(conf.logger)

    # Add the configuration to the run if the logger allows
    if hasattr(logger, "add_config"):
        logger.add_config(OmegaConf.to_container(conf, resolve=True))

    # Instantiating the data
    print("Instantiating the Data")

    # Train (mandatory)
    data = instantiate(conf.data, _convert_="all")

    # Validation (optional)
    if "valid_data" in conf:
        valid_data = instantiate(conf.valid_data, _convert_="all")
    else:
        valid_data = None

    # Test (optional)
    if "test_data" in conf:
        test_data = instantiate(conf.test_data, _convert_="all")
    else:
        test_data = None

    # If the data is sampled from a distribution, compute the true mutual informaiton
    if "distribution" in conf:
        dist = instantiate(conf.distribution, _convert_="all")
        true_mi = dist.mutual_information()
    else:
        true_mi = None

    # Handle estimating mutual information between variables different from 'x' and 'y'
    if "x_key" in conf:
        x_key = conf.x_key
    else:
        x_key = None
    if "y_key" in conf:
        y_key = conf.y_key
    else:
        y_key = None

    # Add the sampled x and y as parameters of the quantization scheme
    # TODO this does not work for multi-estimators and when datasets are passed
    extra_params = {}
    if hasattr(conf.mi_estimator, "quantize_x"):
        print(f"Training the quantization for {x_key}")
        extra_params["quantize_x"] = instantiate(
            conf.mi_estimator.quantize_x,
            _convert_="all",
            data=data[x_key],
        )
    if hasattr(conf.mi_estimator, "quantize_y"):
        print("Training the quantization for y")
        extra_params["quantize_y"] = instantiate(
            conf.mi_estimator.quantize_y,
            _convert_="all",
            data=data[y_key],
        )

    # Instantiate the estimator
    print("Instantiating the Mutual Information Estimator")

    estimator = _instantiate_estimator(
        instantiate,
        data=data,
        config=conf.mi_estimator,
        x_key=x_key,
        y_key=y_key,
        verbose=True,
        _convert_="all",
        **extra_params,
    )

    n_parameters = 0
    for param in estimator.parameters():
        n_parameters += param.numel()
    print(f"{n_parameters} Parameters")

    # Define which quantities are logged
    eval_logged_methods = [
        ("log_ratio", compute_mean_std),
        ("batch_loss", compute_mean_std),
    ]

    if isinstance(estimator, DiscriminativeMIEstimator):
        eval_logged_methods += [
            ("approx_log_partition", compute_mean_std),
            ("unnormalized_log_ratio", compute_mean_std),
        ]

    if isinstance(estimator, HybridMIEstimator):
        eval_logged_methods += [
            ("generative_estimator.log_ratio", compute_mean_std),
            ("generative_estimator.batch_loss", compute_mean_std),
            ("discriminative_estimator.batch_loss", compute_mean_std),
            ("discriminative_estimator.log_ratio", compute_mean_std),
            (
                "discriminative_estimator.approx_log_partition",
                compute_mean_std,
            ),
            (
                "discriminative_estimator.unnormalized_log_ratio",
                compute_mean_std,
            ),
        ]

    train_logged_methods = [("batch_loss", compute_mean_std)]

    # Move back to the output directory created by hydra
    os.chdir(hydra_wd)

    # Train the estimator and evaluate the mutual information
    estimated_mi, train_log = estimate_mi(
        estimator=estimator,
        data=data,
        valid_data=valid_data,
        test_data=test_data,
        logger=logger,
        train_logged_methods=train_logged_methods,
        eval_logged_methods=eval_logged_methods,
        return_estimator=False,
        **conf.estimation,
    )

    if not (true_mi is None):
        print(f"True mi: {true_mi}")
    print(f"Estimated mi: {estimated_mi}")

    logger.save_log()

    if conf.save_trained_model:
        logger.save_model(estimator, "mi_estimator.pyt")


def run():
    OmegaConf.register_new_resolver("eval", eval)
    parse()


if __name__ == "__main__":
    run()



from typing import Optional, Union, Tuple, Dict, Callable, List

import numpy as np
import pandas as pd
import torch
from torch.utils.data import DataLoader, random_split, Dataset

from torch_mist.estimators import MIEstimator, TransformedMIEstimator
from torch_mist.estimators.hybrid import PQHybridMIEstimator
from torch_mist.utils.batch import unfold_samples, move_to_device
from torch_mist.utils.data import SampleDataset, SameAttributeDataLoader
from torch_mist.utils.data.dataset import DataFrameDataset
from torch_mist.utils.data.loader import sample_same_value


TensorDictLike = Union[
    Tuple[Union[torch.Tensor, np.ndarray], Union[torch.Tensor, np.ndarray]],
    Dict[str, Union[torch.Tensor, np.ndarray]],
    Dataset,
    DataLoader,
    pd.DataFrame,
]


def convert_to_tensor(
    array: Union[torch.Tensor, np.array]
) -> torch.FloatTensor:
    if isinstance(array, np.ndarray):
        array = torch.FloatTensor(array)

    return array


def make_dataset(data: TensorDictLike) -> Dataset:
    # Validate the input combinations
    if isinstance(data, list) or isinstance(data, tuple):
        if len(data) != 2:
            raise ValueError("data should be a tuple of two elements (x, y).")
        x, y = data[0], data[1]

        x = convert_to_tensor(x)
        y = convert_to_tensor(y)

        # Unsqueeze an empty dimension if necesary
        if len(x.shape) == 1:
            x = x.view(-1, 1)
        if len(y.shape) == 1:
            y = y.view(-1, 1)

        dataset = SampleDataset({"x": x, "y": y})
    elif isinstance(data, dict):
        dataset = SampleDataset(data)
    elif isinstance(data, pd.DataFrame):
        dataset = DataFrameDataset(data)
    else:
        dataset = data

    return dataset


def make_default_dataloaders(
    data: TensorDictLike,
    valid_data: Optional[TensorDictLike] = None,
    valid_percentage: float = 0.1,
    batch_size: Optional[int] = None,
    num_workers: int = -1,
) -> Tuple[DataLoader, Optional[DataLoader]]:
    # Make the datasets if necessary
    if isinstance(data, DataLoader):
        train_loader = data
        train_set = None
    else:
        if batch_size is None:
            raise ValueError("Please specify a value for batch_size.")
        train_set = make_dataset(data)
        train_loader = None

    if valid_data is None:
        valid_set = None
        valid_loader = None
    elif isinstance(valid_data, DataLoader):
        valid_loader = data
        valid_set = None
    else:
        valid_set = make_dataset(valid_data)
        valid_loader = None

    # Create a validation set if specified
    if valid_percentage > 0:
        if valid_set is None and valid_loader is None:
            if train_set is None:
                raise ValueError(
                    "Please use a tuple (x,y), a dictionary {'x':..,'y':...} or a Dataset instead of a DataLoader"
                    + " for valid_percentage>0. Alternatively, set valid_percentage=0 or specify valid_data."
                )
            else:
                # Make a random train/valid split
                n_valid = int(len(train_set) * valid_percentage)
                train_set, valid_set = random_split(
                    train_set, [len(train_set) - n_valid, n_valid]
                )
        else:
            print(
                "[Warning]: valid_percentage is ignored since valid_set or valid_loader are already specified."
            )

    if train_loader is None:
        train_loader = DataLoader(
            train_set,
            batch_size=batch_size,
            shuffle=True,
            num_workers=num_workers,
        )

    if valid_loader is None and not (valid_set is None):
        valid_loader = DataLoader(
            valid_set, batch_size=batch_size, num_workers=num_workers
        )

    assert isinstance(train_loader, DataLoader) and not (train_loader is None)
    assert isinstance(valid_loader, DataLoader) or (valid_loader is None)

    return train_loader, valid_loader


def modify_data_loader(
    data_loader: DataLoader,
    neg_samples: int,
    device: torch.device,
    transforms: List[Dict[str, Callable]],
) -> DataLoader:
    def compute_attributes(samples) -> torch.Tensor:
        variables = unfold_samples(samples)
        variables = move_to_device(variables, device)
        assert "y" in variables
        y = variables["y"]
        for transform in transforms:
            y = transform["y->y"](y)
        return y.data.cpu()

    if not isinstance(data_loader, SameAttributeDataLoader):
        data_loader = sample_same_value(
            data_loader, compute_attributes, neg_samples=neg_samples
        )

    return data_loader


def make_dataloaders(
    estimator: MIEstimator,
    data: TensorDictLike,
    valid_data: Optional[TensorDictLike] = None,
    valid_percentage: float = 0.1,
    batch_size: Optional[int] = None,
    num_workers: int = 0,
) -> Tuple[DataLoader, Optional[DataLoader]]:
    train_loader, valid_loader = make_default_dataloaders(
        data=data,
        valid_data=valid_data,
        valid_percentage=valid_percentage,
        batch_size=batch_size,
        num_workers=num_workers,
    )

    # Check if we need to modify the dataloader to sample the same attributes
    _estimator = estimator
    transforms = []
    while isinstance(_estimator, TransformedMIEstimator):
        transforms.append(_estimator.transforms)
        _estimator = estimator.base_estimator

    # If required, change the data-loader to sample batches with the same attribute only
    if isinstance(_estimator, PQHybridMIEstimator):
        neg_samples = _estimator.neg_samples
        device = next(iter(estimator.parameters())).device
        transforms.append({"y->y": _estimator.quantize_y})
        train_loader = modify_data_loader(
            train_loader,
            transforms=transforms,
            device=device,
            neg_samples=neg_samples,
        )

        if valid_loader:
            valid_loader = modify_data_loader(
                valid_loader,
                transforms=transforms,
                device=device,
                neg_samples=neg_samples,
            )

    return train_loader, valid_loader



from typing import Type, Optional, Dict, Any, Union, Tuple, List, Callable

import numpy as np
import pandas as pd
import torch
from torch.optim import Optimizer
from tqdm.autonotebook import tqdm
from torch.optim import Adam
from torch.optim.lr_scheduler import (
    SequentialLR,
    LinearLR,
    CosineAnnealingLR,
    ConstantLR,
    LRScheduler,
)
from torch.utils.data import DataLoader, Dataset

from torch_mist.estimators.base import MIEstimator
from torch_mist.utils.batch import unfold_samples, move_to_device

from torch_mist.utils.evaluation import evaluate_mi
from torch_mist.utils.logging import PandasLogger
from torch_mist.utils.logging.logger.base import Logger, DummyLogger
from torch_mist.utils.misc import make_dataloaders, TensorDictLike
from torch_mist.utils.train.utils import RunTerminationManager


def _instantiate_optimizer(
    estimator: MIEstimator,
    max_iterations: int,
    warmup_iterations: int = 0,
    optimizer_class: Type[Optimizer] = Adam,
    optimizer_params: Optional[Dict[str, Any]] = None,
    lr_annealing: bool = False,
) -> Tuple[Optimizer, LRScheduler]:
    params = [
        {"params": params}
        for params in estimator.parameters()
        if params.requires_grad
    ]

    if optimizer_params is None:
        optimizer_params = {"lr": 5e-4}

    opt = optimizer_class(params, **optimizer_params)

    # Cosine annealing with initial linear warmup
    if lr_annealing:
        lr_scheduler = SequentialLR(
            opt,
            [
                LinearLR(
                    opt, start_factor=1e-2, total_iters=warmup_iterations
                ),
                CosineAnnealingLR(
                    opt,
                    T_max=max_iterations - warmup_iterations,
                    eta_min=1e-5,
                ),
            ],
            milestones=[warmup_iterations],
        )
    else:
        lr_scheduler = ConstantLR(opt, 1.0)

    return opt, lr_scheduler


def train_epoch(
    estimator: MIEstimator,
    train_loader: DataLoader,
    opt: Optimizer,
    lr_scheduler: LRScheduler,
    device: Union[str, torch.device],
    logger: Optional[Logger] = None,
    tqdm_iteration: Optional[tqdm] = None,
    fast_train: bool = False,
    train_logged_methods: Optional[
        List[Union[str, Tuple[str, Callable]]]
    ] = None,
    eval_logged_methods: Optional[
        List[Union[str, Tuple[str, Callable]]]
    ] = None,
    max_iterations: Optional[int] = None,
):
    with logger.train():
        estimator.train()
        with logger.epoch():
            for samples in train_loader:
                variables = unfold_samples(samples)
                variables = move_to_device(variables, device)

                if max_iterations:
                    if logger._iteration >= max_iterations:
                        break

                with logger.iteration():
                    with logger.logged_methods(
                        estimator, train_logged_methods
                    ):
                        loss = estimator(**variables)

                    # Compute the ratio only if necessary
                    if not fast_train and not isinstance(logger, DummyLogger):
                        with logger.logged_methods(
                            estimator, eval_logged_methods
                        ):
                            estimator.mutual_information(**variables)

                    opt.zero_grad()
                    loss.backward()
                    opt.step()
                    lr_scheduler.step()

                if tqdm_iteration:
                    tqdm_iteration.update(1)
                    tqdm_iteration.set_postfix_str(f"loss: {loss}")


def validate(
    estimator: MIEstimator,
    valid_loader: Optional[DataLoader],
    device: Union[str, torch.device],
    logger: Logger,
    eval_logged_methods: Optional[
        List[Union[str, Tuple[str, Callable]]]
    ] = None,
) -> float:
    if eval_logged_methods is None:
        eval_logged_methods = []

    if valid_loader is not None:
        with logger.valid():
            with logger.logged_methods(
                estimator,
                eval_logged_methods,
            ):
                valid_mi = evaluate_mi(
                    estimator=estimator,
                    data=valid_loader,
                    device=device,
                )

            if isinstance(valid_mi, dict):
                valid_mi = sum(valid_mi.values())
    else:
        valid_mi = None

    return valid_mi


def train_mi_estimator(
    estimator: MIEstimator,
    data: TensorDictLike,
    valid_data: Optional[TensorDictLike] = None,
    valid_percentage: float = 0.1,
    batch_size: Optional[int] = None,
    num_workers: int = 0,
    device: Union[torch.device, str] = torch.device("cpu"),
    max_epochs: Optional[int] = None,
    max_iterations: Optional[int] = None,
    optimizer_class: Type[Optimizer] = Adam,
    optimizer_params: Optional[Dict[str, Any]] = None,
    lr_annealing: bool = False,
    warmup_percentage: float = 0.2,
    verbose: bool = True,
    logger: Optional[Union[Logger, bool]] = None,
    early_stopping: bool = False,
    patience: int = 5,
    delta: float = 0.001,
    fast_train: bool = False,
    train_logged_methods: Optional[
        List[Union[str, Tuple[str, Callable]]]
    ] = None,
    eval_logged_methods: Optional[
        List[Union[str, Tuple[str, Callable]]]
    ] = None,
) -> Optional[Any]:
    # Create the training and validation dataloaders
    train_loader, valid_loader = make_dataloaders(
        estimator=estimator,
        data=data,
        valid_data=valid_data,
        valid_percentage=valid_percentage,
        batch_size=batch_size,
        num_workers=num_workers,
    )

    if early_stopping and (valid_loader is None):
        print(
            "[Warning]: Please specify a validation set or use valid_percentage>0 to for early_stopping."
        )
        early_stopping = False

    if max_iterations is None and max_epochs is None:
        raise ValueError("Please specify either max_epochs or max_iterations")

    iterations_per_epoch = len(train_loader)

    if max_epochs is None:
        max_epochs = int(np.ceil(max_iterations / iterations_per_epoch))

    if max_iterations is None:
        max_iterations = iterations_per_epoch * max_epochs

    if not 0 <= warmup_percentage <= 1:
        raise ValueError("Warmup percentage must be between 0 and 1")

    warmup_iterations = int(
        iterations_per_epoch * max_epochs * warmup_percentage
    )

    opt, lr_scheduler = _instantiate_optimizer(
        estimator=estimator,
        optimizer_class=optimizer_class,
        optimizer_params=optimizer_params,
        lr_annealing=lr_annealing,
        warmup_iterations=warmup_iterations,
        max_iterations=max_iterations,
    )
    estimator = estimator.to(device)

    default_logger = False
    # If the logger is None, use the default PandasLogger,
    if logger is None:
        logger = PandasLogger()
        default_logger = True
    # If False, instantiate a DummyLogger, which does not store any quantity
    elif logger is False:
        logger = DummyLogger()
        default_logger = True

    # If nothing is specified, log the loss and mutual information
    if train_logged_methods is None:
        train_logged_methods = ["loss"]
    if eval_logged_methods is None:
        eval_logged_methods = ["mutual_information"]

    run_manager = RunTerminationManager(
        early_stopping=early_stopping,
        delta=delta,
        patience=patience,
        verbose=verbose,
        warmup_iterations=warmup_iterations,
        max_iterations=max_iterations,
        maximize=estimator.lower_bound,
        minimize=estimator.upper_bound,
    )

    tqdm_epochs = (
        tqdm(total=max_epochs, desc="Epoch", position=1) if verbose else None
    )
    tqdm_iteration = (
        tqdm(total=len(train_loader), desc="Iteration", position=1)
        if verbose
        else None
    )

    for epoch in range(max_epochs):
        if tqdm_epochs:
            tqdm_iteration.reset()

        train_epoch(
            estimator=estimator,
            train_loader=train_loader,
            opt=opt,
            lr_scheduler=lr_scheduler,
            logger=logger,
            tqdm_iteration=tqdm_iteration,
            device=device,
            fast_train=fast_train,
            train_logged_methods=train_logged_methods,
            eval_logged_methods=eval_logged_methods,
            max_iterations=max_iterations,
        )

        valid_mi = validate(
            estimator=estimator,
            valid_loader=valid_loader,
            device=device,
            logger=logger,
            eval_logged_methods=eval_logged_methods,
        )

        if tqdm_epochs:
            if valid_mi:
                tqdm_epochs.set_postfix_str(f"valid_mi: {valid_mi}")
            tqdm_epochs.update(1)

        if run_manager.should_stop(
            iteration=logger._iteration, valid_mi=valid_mi, model=estimator
        ):
            break

    log = logger.get_log()

    if default_logger:
        logger.clear()

    # Load the state dictionary for the best score
    if not (run_manager.best_state_dict is None):
        estimator.load_state_dict(run_manager.best_state_dict)

    return log



import inspect
from typing import Optional, Any, Union, Type, Dict, Tuple, List, Callable

import torch
import numpy as np
import pandas as pd
from torch.optim import Optimizer, Adam
from torch.utils.data import DataLoader, Dataset

from torch_mist.estimators import MultiMIEstimator
from torch_mist.estimators.base import MIEstimator
from torch_mist.estimators.factories import instantiate_estimator
from torch_mist.utils.dims import infer_dims
from torch_mist.utils.logging import PandasLogger
from torch_mist.utils.logging.logger.base import Logger, DummyLogger
from torch_mist.utils.misc import TensorDictLike
from torch_mist.utils.train.mi_estimator import train_mi_estimator
from torch_mist.utils.evaluation import evaluate_mi


def _instantiate_estimator(
    instantiation_func: Callable[[Any], MIEstimator],
    data: Union[
        Tuple[
            Union[torch.Tensor, np.ndarray], Union[torch.Tensor, np.ndarray]
        ],
        Dict[str, Union[torch.Tensor, np.ndarray]],
        Dataset,
        DataLoader,
    ],
    x_key: Optional[str] = None,
    y_key: Optional[str] = None,
    verbose: bool = True,
    **kwargs,
) -> MIEstimator:
    dims = infer_dims(data)

    if "x" in dims:
        x_dim = dims["x"]
    else:
        if x_key in dims:
            x_dim = dims[x_key]
        else:
            raise ValueError(
                "The data does not contain a key for 'x'.\n"
                + f"Please specify a value for x_key among {dims.keys()}"
            )

    if "y" in dims:
        y_dim = dims["y"]
    else:
        if y_key in dims:
            y_dim = dims[y_key]
        else:
            raise ValueError(
                "The data does not contain a key for 'y'.\n"
                + f"Please specify a value for y_key among {dims.keys()}"
            )

    kwargs["x_dim"] = x_dim
    kwargs["y_dim"] = y_dim

    if verbose:
        print(f"Instantiating the estimator with {kwargs}")

    estimator = instantiation_func(
        **kwargs,
    )

    if verbose:
        print(estimator)

    return estimator


def estimate_mi(
    data: TensorDictLike,
    estimator_name: Optional[str] = None,
    estimator: Optional[MIEstimator] = None,
    valid_data: Optional[TensorDictLike] = None,
    test_data: Optional[TensorDictLike] = None,
    valid_percentage: float = 0.1,
    device: Union[torch.device, str] = torch.device("cpu"),
    max_epochs: Optional[int] = None,
    max_iterations: Optional[int] = None,
    optimizer_class: Type[Optimizer] = Adam,
    optimizer_params: Optional[Dict[str, Any]] = None,
    verbose: bool = True,
    logger: Optional[Union[Logger, bool]] = None,
    lr_annealing: bool = False,
    warmup_percentage: float = 0.2,
    batch_size: Optional[int] = 128,
    evaluation_batch_size: Optional[int] = None,
    num_workers: int = 0,
    early_stopping: bool = True,
    patience: int = 5,
    delta: float = 0.001,
    return_estimator: bool = False,
    fast_train: bool = False,
    hidden_dims: List[int] = [128, 64],
    x_key: Optional[str] = None,
    y_key: Optional[str] = None,
    **kwargs,
) -> Union[
    float,
    Tuple[float, pd.DataFrame],
    Tuple[float, MIEstimator],
    Tuple[float, MIEstimator, pd.DataFrame],
]:
    if max_epochs is None and max_iterations is None:
        if verbose:
            print(
                "max_epochs and max_iterations are not specified, using max_epochs=10 by default."
            )
        max_epochs = 10

    if estimator is None and estimator_name is None:
        raise ValueError(
            "Please specify a value for estimator or estimator_name."
        )

    if estimator is None:
        # Instantiate the estimator while inferring the size for x and y
        if verbose:
            print(f"Instantiating the {estimator_name} estimator")

        estimator = _instantiate_estimator(
            instantiation_func=instantiate_estimator,
            data=data,
            x_key=x_key,
            y_key=y_key,
            verbose=verbose,
            estimator_name=estimator_name,
            hidden_dims=hidden_dims,
            **kwargs,
        )

    # If using different key instead of 'x' and 'y'
    if not (x_key is None) or not (y_key is None):
        if x_key is None:
            x_key = "x"
        if y_key is None:
            y_key = "y"

        estimator = MultiMIEstimator({(x_key, y_key): estimator})

    if verbose:
        print("Training the estimator")

    if evaluation_batch_size is None:
        evaluation_batch_size = batch_size

    if logger is None:
        logger = PandasLogger()
    elif logger is False:
        logger = DummyLogger()

    train_log = train_mi_estimator(
        estimator=estimator,
        data=data,
        valid_data=valid_data,
        valid_percentage=valid_percentage,
        device=device,
        max_epochs=max_epochs,
        max_iterations=max_iterations,
        optimizer_class=optimizer_class,
        optimizer_params=optimizer_params,
        verbose=verbose,
        logger=logger,
        lr_annealing=lr_annealing,
        warmup_percentage=warmup_percentage,
        batch_size=batch_size,
        early_stopping=early_stopping,
        patience=patience,
        delta=delta,
        num_workers=num_workers,
        fast_train=fast_train,
    )

    if verbose:
        print("Evaluating the value of Mutual Information")

    if test_data is None:
        print(
            "[Warning]: using the train_data to estimate the value of mutual information. Please specify test_data."
        )
        test_data = data

    with logger.test():
        mi_value = evaluate_mi(
            estimator=estimator,
            data=test_data,
            batch_size=evaluation_batch_size,
            device=device,
            num_workers=num_workers,
        )

    logger.clear()

    if not (train_log is None) and return_estimator:
        return mi_value, estimator, train_log
    elif not (train_log is None):
        return mi_value, train_log
    elif return_estimator:
        return mi_value, estimator
    else:
        return mi_value



from typing import Union, Dict, Tuple

import torch
import numpy as np
from torch.utils.data import DataLoader, Dataset

from torch_mist.utils.batch import unfold_samples
from torch_mist.utils.misc import make_default_dataloaders


def infer_dims(
    data: Union[
        Tuple[
            Union[torch.Tensor, np.ndarray], Union[torch.Tensor, np.ndarray]
        ],
        Dict[str, Union[torch.Tensor, np.ndarray]],
        Dataset,
        DataLoader,
    ],
) -> Dict[str, int]:
    dataloader, _ = make_default_dataloaders(
        data=data,
        valid_data=None,
        batch_size=1,
        valid_percentage=0,
        num_workers=0,
    )

    batch = next(iter(dataloader))
    variables = unfold_samples(batch)
    return {k: v.shape[-1] for k, v in variables.items()}



from typing import Optional

from torch import nn


class RunTerminationManager:
    def __init__(
        self,
        early_stopping: bool,
        delta: float,
        patience: int,
        verbose: bool,
        warmup_iterations: Optional[int],
        max_iterations: Optional[int],
        maximize: bool = False,
        minimize: bool = False,
    ):
        if early_stopping:
            if not maximize and not minimize:
                print(
                    "[Warning]: early_stopping can be used only when maximizing or minimizing, the parameter will be ignored."
                )
                early_stopping = False
        self.early_stopping = early_stopping
        self.maximize = maximize
        self.minimize = minimize
        self.best_value = 0
        self.delta = delta
        self.patience = patience
        self.current_patience = patience
        self.verbose = verbose
        self.max_iterations = max_iterations
        self.warmup_iterations = warmup_iterations
        self.best_state_dict = None

    def should_stop(
        self, iteration: int, valid_mi: Optional[float], model: nn.Module
    ) -> bool:
        stop = False

        if iteration <= self.warmup_iterations:
            return False

        if self.early_stopping:
            improvement = (
                (valid_mi - self.best_value)
                if self.maximize
                else (self.best_value - valid_mi)
            )
            if improvement >= self.delta:
                # Improvement, update best and reset the patience
                self.best_value = valid_mi
                self.current_patience = self.patience
                self.best_state_dict = model.state_dict()
            else:
                self.current_patience -= 1
                if self.verbose:
                    print(f"Loosing patience: {self.current_patience}")

            if self.current_patience <= 0:
                if self.verbose:
                    print("No improvements on validation, stopping.")
                stop = True

        if self.max_iterations:
            if iteration >= self.max_iterations:
                stop = True
        return stop
